// generated by verigenSV.pl on Fri Jun  4 13:21:36 2021
// input file: mesq.cfg

// compiler directives
/* verilator lint_off CASEINCOMPLETE */
/* verilator lint_off DECLFILENAME */
/* verilator lint_off UNUSED */
/* verilator lint_off STMTDLY */
/* verilator lint_off UNOPTFLAT */


module MESQ (
	input logic clk,
	input logic res_n,
	input logic opc_vld,
	input logic [OPC_WIDTH-1:0] opc,
	input logic rd_q0,
	input logic rd_q1,
	output logic [OPC_WIDTH-1:0] data_q0, 
	output logic empty_q0, 
	output logic full_q0, 
	output logic [OPC_WIDTH-1:0] data_q1, 
	output logic empty_q1, 
	output logic full_q1 );

  //
 // parameter constants 
//---------------------------
parameter int OPC_WIDTH = 5 ;
parameter int QUEUE_SIZE = 8 ;
parameter logic [OPC_WIDTH-1:0] ZERO_DATA = {OPC_WIDTH{1'b0}};


  //
 // state encoding typedef 
//--------------------------------
typedef enum logic [0:0] {
	Q0 = 1'b0,
	Q1 = 1'b1
} states_t;


// internal state variables
states_t STATE, NEXT_STATE;

  //
 // memory instantiations 
//---------------------------
QUEUE #( OPC_WIDTH, QUEUE_SIZE ) Q0_QUEUE(
	.clk(clk), .res_n(res_n), 
	.wr(Q0_QUEUE_wr), .wr_data(Q0_QUEUE_wr_data[OPC_WIDTH-1:0]), 
	.rd(Q0_QUEUE_rd), .rd_data(Q0_QUEUE_rd_data[OPC_WIDTH-1:0]), 
	.empty(empty_q0), .full(full_q0) );

QUEUE #( OPC_WIDTH, QUEUE_SIZE ) Q1_QUEUE(
	.clk(clk), .res_n(res_n), 
	.wr(Q1_QUEUE_wr), .wr_data(Q1_QUEUE_wr_data[OPC_WIDTH-1:0]), 
	.rd(Q1_QUEUE_rd), .rd_data(Q1_QUEUE_rd_data[OPC_WIDTH-1:0]), 
	.empty(empty_q1), .full(full_q1) );

   //
  // implicit continuous assignments 
 // (derived from <expressions> section) 
//----------------------------------
uwire IDLE; assign IDLE = ~opc_vld & res_n;
uwire WRITE_QUEUE; assign WRITE_QUEUE = opc_vld & res_n;

  //
 // initials for simulator
//---------------------------
initial begin
  STATE = Q0;

  // dump output (Verilator)
  if ($test$plusargs("trace") != 0) begin
    $display("[%0t] Tracing to logs/vlt_dump.vcd...\n", $time);
    $dumpfile("logs/vlt_dump.vcd");
    $dumpvars();
  end
  $display("[%0t] Model running...\n", $time);

end

  //
 // state transfer
//---------------------------
always_ff @( posedge clk ) begin
  if( res_n )
    STATE <= NEXT_STATE;
  else
    STATE <= Q0;

  // condition for $finish (Verilator)
  if( rd_q0==1'b1 && rd_q1==1'b1 && res_n==1'b0 ) begin
    $write("*-* All Finished *-*\n");
    $finish;
  end   

end

// AssertionExample:
//always_ff @ (posedge clk) begin
//   AssertionExample: assert (rd==1'b1 ^ wr==1'b1);
//end

// coverage analysis (Verilator)
cover property ( @(posedge clk) rd_q0==1'b1 && rd_q1==1'b1 && res_n==1'b0 );

always_comb begin

    //
   // pre-case (default) assignments 
  //----------------------------------
  data_q0=POP_Q0_QUEUE(rd_q0);
  data_q1=POP_Q1_QUEUE(rd_q1);

  unique case ( STATE )
  Q0 : begin
	if ( IDLE ) begin
	  NEXT_STATE = Q0;
	end
	if ( WRITE_QUEUE ) begin
	  NEXT_STATE = Q1;
	  PUSH_Q0_QUEUE(1'b1,opc);
	  NOP_Q1_QUEUE();
	end
  end // Q0
  Q1 : begin
	if ( IDLE ) begin
	  NEXT_STATE = Q1;
	end
	if ( WRITE_QUEUE ) begin
	  NEXT_STATE = Q0;
	  PUSH_Q1_QUEUE(1'b1,opc);
	  NOP_Q0_QUEUE();
	end
  end // Q1
  endcase
end // always_comb


  //
 // memory functions
//---------------------------
logic Q0_QUEUE_rd;
logic [OPC_WIDTH-1:0] Q0_QUEUE_rd_data;

function automatic [OPC_WIDTH-1:0] POP_Q0_QUEUE( read );
   Q0_QUEUE_rd = read;
   return Q0_QUEUE_rd_data;
endfunction

logic Q0_QUEUE_wr;
logic [OPC_WIDTH-1:0] Q0_QUEUE_wr_data;

function automatic void PUSH_Q0_QUEUE( write, [OPC_WIDTH-1:0] write_data);
   Q0_QUEUE_wr = write;
   Q0_QUEUE_wr_data = write_data;
endfunction

function automatic void NOP_Q0_QUEUE();
   Q0_QUEUE_rd = 0;
   Q0_QUEUE_wr = 0;
   Q0_QUEUE_wr_data = '0;
endfunction

logic Q1_QUEUE_rd;
logic [OPC_WIDTH-1:0] Q1_QUEUE_rd_data;

function automatic [OPC_WIDTH-1:0] POP_Q1_QUEUE( read );
   Q1_QUEUE_rd = read;
   return Q1_QUEUE_rd_data;
endfunction

logic Q1_QUEUE_wr;
logic [OPC_WIDTH-1:0] Q1_QUEUE_wr_data;

function automatic void PUSH_Q1_QUEUE( write, [OPC_WIDTH-1:0] write_data);
   Q1_QUEUE_wr = write;
   Q1_QUEUE_wr_data = write_data;
endfunction

function automatic void NOP_Q1_QUEUE();
   Q1_QUEUE_rd = 0;
   Q1_QUEUE_wr = 0;
   Q1_QUEUE_wr_data = '0;
endfunction

endmodule


  //
 // memory
//---------------------------
module QUEUE ( clk, res_n, wr, wr_data, rd, rd_data, empty, full );

parameter DATA_WIDTH = 32;
parameter NUM_ENTRIES = 16;
parameter ADDR_WIDTH = $clog2(NUM_ENTRIES);

input  logic clk, res_n, rd, wr;
input  logic [DATA_WIDTH-1:0] wr_data;
output logic [DATA_WIDTH-1:0] rd_data;
output logic empty, full;
    
parameter ZERO_ADDR = {ADDR_WIDTH{1'b0}};
parameter ZERO_DATA = {DATA_WIDTH{1'b0}};

logic [DATA_WIDTH-1:0] MEMORY [NUM_ENTRIES-1:0];
logic [ADDR_WIDTH-1:0] HEAD; 
logic [ADDR_WIDTH-1:0] TAIL; 

integer i;

initial begin
    for ( i = 0; i < NUM_ENTRIES; i = i + 1 ) begin
        MEMORY[ i ] = ZERO_DATA;
    end
    HEAD = ZERO_ADDR;
    TAIL = ZERO_ADDR;
end

// IN-CYCLE READ ACCESS
assign rd_data = ( rd && res_n ) ? MEMORY[HEAD] : ZERO_DATA;
assign empty = (HEAD==TAIL);
assign full = (HEAD==(TAIL+1'b1));

always_ff @ ( posedge clk ) begin
    if( res_n ) begin
        if( rd & !empty ) HEAD <= (HEAD + 1'b1); // READ ACCESS
        if( wr & !full ) begin // WRITE ACCESS
	    MEMORY[TAIL] <= wr_data;
            TAIL <= (TAIL + 1'b1);
        end
    end else begin
	HEAD <= ZERO_ADDR;
        TAIL <= ZERO_ADDR;
        MEMORY[ZERO_ADDR] <= ZERO_DATA;
    end
end

endmodule
