#include <iostream>
#include <verilated.h>
#include "VMESQ.h"

using namespace std;

// Current simulation time (64-bit unsigned)
vluint64_t main_time = 0;
vluint64_t max_time = 100;

// Called by $time in Verilog
double sc_time_stamp() {
    return main_time;  // Note does conversion to real, to match SystemC
}

int main(int argc, char** argv, char** env){
  
  // Prevent unused variable warnings
  if (false && argc && argv && env) {}

  // Set debug level, 0 is off, 9 is highest presently used
  // May be overridden by commandArgs
  Verilated::debug(0);

  // Randomization reset policy
  // May be overridden by commandArgs
  Verilated::randReset(2);

  // Verilator must compute traced signals
  Verilated::traceEverOn(true);

  // Pass arguments so Verilated code can see them, e.g. $value$plusargs
  // This needs to be called before you create any model
  Verilated::commandArgs(argc, argv);

  // Create logs/ directory in case we have traces to put under it
  Verilated::mkdir("logs");

  // Construct the Verilated model, from Vtop.h generated from Verilating "top.v"
  VMESQ* top = new VMESQ();

  top->clk = 0;
  top->res_n = !0;
  
  top->rd_q0 = 0;
  top->rd_q1 = 0;

  top->opc_vld = 0;
  top->opc = 0b00000;
    
  while(!Verilated::gotFinish()){

    main_time++;

    // toggle clock
    top->clk=!top->clk;

    if(!top->clk){
    
      // Toggle control signals on an edge that doesn't correspond
      // to where the controls are sampled; in this example we do
      // this only on a negedge of clk, because we know
      // reset is not sampled there.
      if ( main_time > 0 && main_time <= 10 ) {
	top->res_n = !1;  //   Assert reset
      } else {
	top->res_n = !0;  // Deassert reset
      }

      if( main_time>10 && main_time<=20 ){
	// alternating reads on empty queue
	top->opc_vld = 0;
	if (25==main_time) {
	  top->rd_q0 = 0;
	  top->rd_q1 = 0;
	}
	if (30==main_time) {
	  top->rd_q0 = 0;
	  top->rd_q1 = 0;
	}
      } else {
	if (main_time>20 && main_time<=floor(max_time/2)) {
	  // write on empty queue
	  top->rd_q0 = 0;
	  top->rd_q1 = 0;
	  // increment or wrap opc
	  if (0==main_time%2) {
	    if(top->opc==0b11111) {
	      top->opc=0b00000;
	    } else {
	      top->opc += 0x1;
	    }
	  }
	  // write opc one cycle later
	  if (0==main_time%2 ) {
	    top->opc_vld = 1;	    
	  } else {
	    top->opc_vld = 0;
	  }
	} else {
	  // read on full queue
	  top->opc_vld = 0;
	  if (0==main_time%4) {
	    top->rd_q0 = 1;
	    top->rd_q1 = 0;
	  } else {
	    top->rd_q0 = 0;
	    top->rd_q1 = 1;
	  }
	}
      }

      if(main_time>=max_time){
	top->rd_q0=1; top->rd_q1=1; top->res_n=0;
      }
	
    }

       
    // Evaluate model
    // (If you have multiple models being simulated in the same
    // timestep then instead of eval(), call eval_step() on each, then
    // eval_end_step() on each.)
    top->eval();
    
    // Read outputs
    VL_PRINTF( "[%" VL_PRI64 "d] clk=%x res_n=%x opc=%02x opc_vld=%x ( rd_q0=%x, data_q0=%02x ) ( rd_q1=%x, data_q1=%02x )\n",
	       main_time,
	       top->clk, top->res_n,
	       top->opc, top->opc_vld,
	       top->rd_q0, top->data_q0,
	       top->rd_q1, top->data_q1 );             
  }
  
  top->final();

  //  Coverage analysis (since test passed)
#if VM_COVERAGE
  Verilated::mkdir("logs");
  VerilatedCov::write("logs/coverage.dat");
#endif

  // Delete model
  delete top;
  top = nullptr;

  // Exit
  exit(0);
  
}
